
#include "Manifolds/SymFixedRankE.h"

/*Define the namespace*/
namespace ROPTLIB{

    SymFixedRankE::SymFixedRankE(integer inn, integer inp)
    {
        n = inn;
        p = inp;
        name.assign("SymFixedRankEmbedded");
        IsIntrApproach = false;
//        IsVectTranSmooth = true;
        ExtrinsicDim = integer(n * (n+1) / 2);          // the dimension of symmetric matrices space of size nxn
        IntrinsicDim = integer(n*p - p * (p - 1) / 2);  // the dimension of the manifold.
        EMPTYEXTR = Vector(n*p + p*p, 1);               // EMPTYEXTR = [Z(:); S(:)] of size (n*p+p*p)x1
        EMPTYINTR = Vector(1);     // its dimension is wrong, but we do not use it.
        EMPTYNORINTR = Vector(1);  // its dimension is wrong, but we do not use it.
    };

    SymFixedRankE::~SymFixedRankE()
    {
    };

    realdp SymFixedRankE::Metric(const Variable &x, const Vector &etax, const Vector &xix) const
    {
//        Vector Seta(p, p),  Zeta(n, p);
//        Vector Sxi(p, p),   Zxi(n, p);

//        extractSubMatrix(etax, 0, &Zeta);
//        extractSubMatrix(etax, n*p, &Seta);
//        extractSubMatrix(xix, 0, &Zxi);
//        extractSubMatrix(xix, n*p, &Sxi);

        Vector Zeta = etax.GetSubmatrix(0, n*p-1, 0, 0).Reshape(n, p);
        Vector Seta = etax.GetSubmatrix(n*p, n*p + p*p - 1, 0, 0).Reshape(p, p);
            
        Vector Zxi = xix.GetSubmatrix(0, n*p-1, 0, 0).Reshape(n, p);
        Vector Sxi = xix.GetSubmatrix(n*p, n*p + p*p - 1, 0, 0).Reshape(p, p);
                
        return Seta.DotProduct(Sxi) + 2.*Zeta.DotProduct(Zxi);
    };

    Variable SymFixedRankE::RandominManifold(void) const
    {
        Vector V(n, p), Diag(p, p);
        Diag.SetToZeros();
        V.RandGaussian();
        V = V.GetOrth();
        
        Vector D(p, p);
        D.RandGaussian();
        D.SVDDecom();
        
        Vector result = Vector(n*p + p*p,1);  // we do not use the explicit representation of X
                
        writeMatrix(V, 0, result);
        
        vectorToDiag(D.Field("_S"), Diag);
        
        writeMatrix(Diag, n*p, result);

        return result;
    };

    Variable SymFixedRankE::RandominTxManifold(const Variable &x) const
    {
        Vector V(n, p);
        extractSubMatrix(x, n*p, V);
//        Vector V = x.GetSubmatrix(0, n*p-1, 0, 0).Reshape(n, p);
        Vector S(p, p), Z(n, p);
        S.RandGaussian();
        Z.RandGaussian();
        
        Vector result(x);
        Vector SS = S + S.GetTranspose();
        Vector ZZ = Z - V * ((V.GetTranspose())*Z);
        
        writeMatrix(ZZ, 0, result);
        writeMatrix(SS, n*p, result);
        
        return result;
    };

    Vector &SymFixedRankE::LinearOPEEta(const Variable &x, const LinearOPE &Hx, const Vector &etax, Vector *result) const
    {
        Manifold::LinearOPEEta(x, Hx, etax, result);
        return ExtrProjection(x, *result, result);
    };

    /* result = a etax */
    Vector &SymFixedRankE::ScalarTimesVector(const Variable &x, const realdp &scalar, const Vector &etax, Vector *result) const
    {
        *result = etax;
        result->ScalarTimesThis(scalar);
        
        return *result;
    };


    Vector &SymFixedRankE::ScalarVectorAddVector(const Variable &x, const realdp &scalar, const Vector &etax, const Vector &xix, Vector *result) const
    {

        *result = xix;
        result->AlphaXaddThis(scalar, etax);
        
        return *result;
    };

    /* Compute result = scalar1 * etax + scalar2 * xix */
    Vector &SymFixedRankE::VectorLinearCombination(const Variable &x, realdp scalar1, const Vector &etax, realdp scalar2, const Vector &xix, Vector *result) const
    {
        *result = xix;
        result->ScalarTimesThis(scalar2);
        result->AlphaXaddThis(scalar1, etax);
        
        return *result;
    };

    /* to be done */
    Vector &SymFixedRankE::ObtainNorVerIntr(const Variable &x, const Vector &etax, Vector *result) const
    {
        printf("Warning: FixedRankE::VecTranDiffRetAdjoint has not been overridden!\n");
        assert(IsIntrApproach == true);
        return *result;
    };
    
    /* to be done */
    Vector &SymFixedRankE::ObtainNorVerExtr(const Variable &x, const Vector &intretax, Vector *result) const
    {
        printf("Warning: FixedRankE::VecTranDiffRetAdjoint has not been overridden!\n");
        assert(IsIntrApproach == true);
        return *result;
    };

    /* to be done */
    Variable &SymFixedRankE::Retraction(const Variable &x, const Vector &etax, Variable *result) const
    {
        /* S = x.D + eta.S, Z = eta.Z */
        Vector S(p,p), Z(n,p), Dx(p, p), Vx(n, p);
        extractSubMatrix(etax, 0, Z);
        extractSubMatrix(etax, n*p, S);
        extractSubMatrix(x, 0, Vx);
        extractSubMatrix(x, n*p, Dx);
        
//        const integer np = n*p;
//        const integer npDpp = np + p*p;
//        Vector Z = etax.GetSubmatrix(0, np - 1, 0, 0).Reshape(n, p);
//        Vector S = etax.GetSubmatrix(np, npDpp - 1, 0, 0).Reshape(p, p);
//        Vector Vx = x.GetSubmatrix(0, np - 1, 0, 0).Reshape(n, p);
//        Vector Dx = x.GetSubmatrix(np, npDpp - 1, 0, 0).Reshape(p, p);

        /* S = x.D + eta.S */
        S = S + Dx;
        /* Z = QR */
        Z.QRDecom();
        Vector M(2*p, 2*p), M1(p, 2*p), M2(p, 2*p);
        M.SetToZeros();
        M2.SetToZeros();
        M2.SetToZeros();
        Vector R = Z.Field("_R");
        Vector Rt = R.GetTranspose();
        Vector Zero(p, p);
        Zero.SetToZeros();
        
        concatenateMatricesHorizontally(S, Rt, S.Getrow(), S.Getcol(), Rt.Getcol(), M1);
        concatenateMatricesHorizontally(R, Zero, R.Getrow(), R.Getcol(), Zero.Getcol(), M2);

        M1.Transpose();
        M2.Transpose();
        
        concatenateMatricesHorizontally(M1, M2, M1.Getrow(), M1.Getcol(), M2.Getcol(), M);
        M.Transpose();
        /* [V1, D1] = svd(M) */
        M.SVDDecom();
        
        Vector V1 = M.Field("_U").GetSubmatrix(0, 2*p-1, 0, p-1);
        Vector D1 = M.Field("_S").GetSubmatrix(0, p-1, 0, 0);
        
        Vector Vp = Z.Field("_Q");
        Vector V(n, 2*p);
        /* V = [x.V, Vp] */
        concatenateMatricesHorizontally(Vx, Vp, Vx.Getrow(), Vx.Getcol(), Vp.Getcol(), V);
        
//        Vector Xtmp(n*p + p*p, 1);
//        *result = Xtmp;
        
        Vector VV1 = V*V1;
        
        writeMatrix(VV1, 0, *result);
        
        Vector Diag(p, p);
        Diag.SetToZeros();
        
        vectorToDiag(D1, Diag);
                        
        writeMatrix(Diag, n*p, *result);
        
        return *result;
    };

    /* to be done */
    Vector &SymFixedRankE::VecTranDiffRet(const Variable &x, const Vector &etax, const Variable &y, const Vector &xix, Vector *result, bool IsEtaXiSameDir) const
    {
        printf("Warning: FixedRankE::VecTranDiffRetAdjoint has not been overridden!\n");
        
        return *result;
    };

    /* to be done */
    Vector &SymFixedRankE::VecTranDiffRetAdjoint(const Variable &x, const Vector &etax, const Variable &y, const Vector &xiy, Vector *result) const
    {
        printf("Warning: FixedRankE::VecTranDiffRetAdjoint has not been overridden!\n");
        return Manifold::VecTranDiffRetAdjoint(x, etax, y, xiy, result);
    };

    /* to be done */
    Vector &SymFixedRankE::InverseVecTranDiffRetAdjoint(const Variable &x, const Vector &etax, const Variable &y, const Vector &xix, Vector *result) const
    {
        printf("Warning: FixedRankE::VecTranDiffRetAdjoint has not been overridden!\n");
        return *result;
    };

    /* to be done */
    Vector &SymFixedRankE::InverseVecTranDiffRet(const Variable &x, const Vector &etax, const Variable &y, const Vector &xiy, Vector *result, bool IsEtaXiSameDir) const
    {
        printf("Warning: FixedRankE::VecTranDiffRetAdjoint has not been overridden!\n");
        return *result;
    };

    
    Vector &SymFixedRankE::Projection(const Variable &x, const Vector &etax, Vector *result) const
    {
        return ExtrProjection(x, etax, result);
    };

    /* correct */
    Vector &SymFixedRankE::ExtrProjection(const Variable &x, const Vector &etax, Vector *result) const
    {
        *result = etax;
        
        Vector V(n, p);
        extractSubMatrix(x, 0, V);
        
        Vector Z(n, p);
        extractSubMatrix(etax, 0, Z);
        
        Vector S(p, p);
        extractSubMatrix(etax, n*p, S);
        
        Vector VZ = V.GetTranspose() * Z;        
        Vector S1 = S + VZ + (VZ.GetTranspose());
        Vector Z1 = Z - V*VZ;
        
        writeMatrix(Z1, 0, *result);
        writeMatrix(S1, n*p, *result);
        
        return *result;
    };

    /* we do not need to convert the grad to Riemannian virson */
    Vector &SymFixedRankE::EucGradToGrad(const Variable &x, const Vector &egf, const Problem *prob, Vector *result)  const
    {
        Projection(x, egf, result);
//        *result = egf;
        return *result;
    };

    /* we do not need to convert the grad to Riemannian virson */
    Vector &SymFixedRankE::EucHvToHv(const Variable &x, const Vector &etax, const Vector &exix, const Problem *prob, Vector *result) const
    {
        Projection(x, exix, result);
//        *result = exix;
        return *result;
    };

/*  functions listed below are useful for constructing Problem and Manifold classes.  */
    Vector &SymFixedRankE::concatenateMatricesHorizontally(const Vector &A, const Vector &B, const integer rows, const integer colsA, const integer colsB, Vector &result)
    {
        realdp *resultptr = result.ObtainWriteEntireData();
        const realdp *Aptr = A.ObtainReadData();
        const realdp *Bptr = B.ObtainReadData();
        // ensure that the number of rows of A is the same as the one of B
        
        // Write the data of A into result, first.
        for (integer i = 0; i < rows*colsA; i++)
        {
            resultptr[i] = Aptr[i];
        }
        // Append data of B into result, second.
        for (integer i = 0; i <rows*colsB; i++)
        {
            resultptr[i + rows*colsA] = Bptr[i];
        }
        return result;
    };


    Vector &SymFixedRankE::diagAddM(const Vector &v, const Vector &B, Vector &result)
    {
        /* ensure that B is a suqare matrix */
        assert(B.Getrow() == B.Getcol());
        const realdp *vptr = v.ObtainReadData();
        const integer rowB = B.Getrow();
        B.CopyTo(result);
        realdp *resultptr = result.ObtainWriteEntireData();
        for (integer i = 0; i < rowB; i++)
        {
            resultptr[i * rowB + i] += vptr[i];
        }
        return result;
    };

    Vector &SymFixedRankE::matrixTimesDiagonalVector(const Vector &A, const Vector &v, const char *side, const char *mord, Vector &result)
    {
        const realdp *Aptr = A.ObtainReadData();
        const realdp *vptr = v.ObtainReadData();
        realdp *resultptr = result.ObtainWriteEntireData();
        
        const integer rowA = A.Getrow();
        const integer colA = A.Getcol();
        const integer rowv = v.Getrow();
        
        realdp diagElement = 0;
        
        for (integer i = 0; i < rowv; i++)
        {
            // get the i-th main element of Diag(v)
            if (strcmp(mord, "m") == 0 || strcmp(mord, "M") == 0)
                diagElement = vptr[i];
            else
                diagElement = 1. / vptr[i];
            if (strcmp(side, "R") == 0 || strcmp(side, "r") == 0)
            {
                assert(colA == rowv);
                // traversing the i-th col of Matrix A, denoted by col(A, i)
                for (int j = 0; j < rowA; j++)
                {
                    // col(result, i) <- v(i) * col(A, i)
                    resultptr[i * rowA + j] = Aptr[i * rowA + j] * diagElement;
                }
            }
            else if (strcmp(side, "L") == 0 || strcmp(side, "l") == 0)
            {
                assert(rowA == rowv);
                // traversing the i-th row of Matrix A, denoted by row(A, i)
                for (int j = 0; j < colA; j++)
                {
                    // row(result, i) <- v(i) * row(A, i)
                    resultptr[i + rowA * j] = Aptr[i + rowA * j] * diagElement;
                }
            }
        }
        
        return  result;
    };

    /* get wrong */
    Vector &SymFixedRankE::extractSubMatrix(const Vector &A, const integer num, Vector &result)
    {
        const integer n = result.Getrow();
        const integer p = result.Getcol();
        
        const realdp *Aptr = A.ObtainReadData();
        realdp *resultptr = result.ObtainWriteEntireData();
        for (integer i = 0; i < n*p; i++)
                resultptr[i] = Aptr[num + i];
        return result;
    };

    /* correct */
    Vector &SymFixedRankE::vectorToDiag(const Vector &v, Vector &result)
    {
        const realdp *vptr = v.ObtainReadData();
        realdp *resultptr = result.ObtainWriteEntireData();
                        
        const integer rowv = v.Getrow();
        
        for (integer i = 0; i < rowv; i++)
            resultptr[i*rowv + i] = vptr[i];
        
        return result;
    };

    /* correct */
    Vector &SymFixedRankE::writeMatrix(const Vector &A, const integer num, Vector &result)
    {
        const realdp *Aptr = A.ObtainReadData();
        realdp *resultptr = result.ObtainWriteEntireData();

        realdp rowxcol = A.Getrow()*A.Getcol();

        for (integer i = 0; i < rowxcol; i++)
            resultptr[i + num] = Aptr[i];

        return result;
    };

}; /*end of ROPTLIB namespace*/
